@@ -1,9 +1,31 @@
+import asyncio
+import enum
 import logging
-import greeclimate.network_helper as nethelper
-
 from enum import IntEnum, unique
-from greeclimate.exceptions import DeviceNotBoundError
-from greeclimate.network_helper import Props
+
+import greeclimate.network as network
+from greeclimate.exceptions import DeviceNotBoundError, DeviceTimeoutError
+
+
+class Props(enum.Enum):
+    POWER = "Pow"
+    MODE = "Mod"
+    TEMP_SET = "SetTem"
+    TEMP_UNIT = "TemUn"
+    TEMP_BIT = "TemRec"
+    FAN_SPEED = "WdSpd"
+    FRESH_AIR = "Air"
+    XFAN = "Blo"
+    ANION = "Health"
+    SLEEP = "SwhSlp"
+    LIGHT = "Lig"
+    SWING_HORIZ = "SwingLfRig"
+    SWING_VERT = "SwUpDn"
+    QUIET = "Quiet"
+    TURBO = "Tur"
+    STEADY_HEAT = "StHt"
+    POWER_SAVE = "SvSt"
+    UNKNOWN_HEATCOOLTYPE = "HeatCoolType"
 
 
 @unique
@@ -58,6 +80,55 @@ class VerticalSwing(IntEnum):
     SwingLower = 11
 
 
+class DeviceInfo:
+    """Device information class, used to identify and connect
+
+    Attributes
+        ip: IP address (ipv4 only) of the physical device
+        port: Usually this will always be 7000
+        mac: mac address, in the format 'aabbcc112233'
+        name: Name of unit, if available
+    """
+
+    ip = None
+    port = None
+    mac = None
+    name = "<No Name>"
+    brand = None
+    model = None
+    version = None
+
+    def __init__(self, ip, port, mac, name, brand=None, model=None, version=None):
+        self.ip = ip
+        self.port = port
+        self.mac = mac
+        self.name = name if name else self.name
+        self.brand = brand
+        self.model = model
+        self.version = version
+
+    def __str__(self):
+        return f"Device: {self.name} @ {self.ip}:{self.port} (mac: {self.mac})"
+
+    def __eq__(self, other):
+        """Check equality based on Device Info properties"""
+        if isinstance(other, DeviceInfo):
+            return (
+                self.ip == other.ip
+                and self.port == other.port
+                and self.mac == other.mac
+                and self.name == other.name
+                and self.brand == other.brand
+                and self.model == other.model
+                and self.version == other.version
+            )
+        return False
+
+    def __ne__(self, other):
+        """Check inequality based on Device Info properties"""
+        return not self.__eq__(other)
+
+
 class Device:
     """Class representing a physical device, it's state and properties.
 
@@ -95,10 +166,11 @@ class Device:
 
         """ Device properties """
         self._properties = None
+        self._dirty = []
 
     async def bind(self, key=None):
-        """ Run the binding procedure.
-        
+        """Run the binding procedure.
+
         Binding is a finnicky procedure, and happens in 1 of 2 ways:
             1 - Without the key, binding must pass the device info structure immediately following
                 the search devices procedure. There is only a small window to complete registration.
@@ -107,18 +179,24 @@ class Device:
             Both approaches result in a device_key which is used as like a persitent session id.
 
         Raises:
-            socket.timeout: If binding was unsuccessful (the device didn't respond.)
+            DeviceNotBoundError: If binding was unsuccessful (the device didn't respond.)
         """
 
+        if not self.device_info:
+            raise DeviceNotBoundError
+
         self._logger.info("Starting device binding to %s", str(self.device_info))
 
-        if key:
-            self.device_key = key
-        else:
-            self.device_key = await nethelper.bind_device(self.device_info)
+        try:
+            if key:
+                self.device_key = key
+            else:
+                self.device_key = await network.bind_device(self.device_info)
 
-        if self.device_key:
-            self._logger.info("Bound to device using key %s", self.device_key)
+            if self.device_key:
+                self._logger.info("Bound to device using key %s", self.device_key)
+        except asyncio.TimeoutError:
+            raise DeviceNotBoundError
 
     async def update_state(self):
         """ Update the internal state of the device structure of the physical device """
@@ -128,9 +206,33 @@ class Device:
         self._logger.debug("Updating device properties for (%s)", str(self.device_info))
 
         props = [x.value for x in Props]
-        self._properties = nethelper.request_state(
-            props, self.device_info, self.device_key
-        )
+
+        try:
+            self._properties = await network.request_state(
+                props, self.device_info, self.device_key
+            )
+        except asyncio.TimeoutError:
+            raise DeviceTimeoutError
+
+    async def push_state_update(self):
+        """ Push any pending state updates to the unit """
+        if not self._dirty:
+            return
+
+        self._logger.debug("Pushing state updates to (%s)", str(self.device_info))
+
+        props = {}
+        for name in self._dirty:
+            value = self._properties.get(name)
+            self._logger.debug("Sending remote state update %s -> %s", name, value)
+            props[name] = value
+
+        self._dirty.clear()
+
+        try:
+            await network.send_state(props, self.device_info, key=self.device_key)
+        except asyncio.TimeoutError:
+            raise DeviceTimeoutError
 
     def get_property(self, name):
         """ Generic lookup of properties tracked from the physical device """
@@ -147,48 +249,47 @@ class Device:
             return
         else:
             self._properties[name.value] = value
-
-        self._logger.debug("Sending remote state update %s -> %s", name, value)
-        nethelper.send_state({name.value: value}, self.device_info, key=self.device_key)
+            if name.value not in self._dirty:
+                self._dirty.append(name.value)
 
     @property
     def power(self) -> bool:
         return bool(self.get_property(Props.POWER))
 
     @power.setter
-    def power(self, value):
+    def power(self, value: int):
         self.set_property(Props.POWER, int(value))
 
     @property
     def mode(self) -> int:
-        return int(self.get_property(Props.MODE))
+        return self.get_property(Props.MODE)
 
     @mode.setter
-    def mode(self, value):
+    def mode(self, value: int):
         self.set_property(Props.MODE, int(value))
 
     @property
     def target_temperature(self) -> int:
-        return int(self.get_property(Props.TEMP_SET))
+        return self.get_property(Props.TEMP_SET)
 
     @target_temperature.setter
-    def target_temperature(self, value):
+    def target_temperature(self, value: int):
         self.set_property(Props.TEMP_SET, int(value))
 
     @property
     def temperature_units(self) -> int:
-        return int(self.get_property(Props.TEMP_UNIT))
+        return self.get_property(Props.TEMP_UNIT)
 
     @temperature_units.setter
-    def temperature_units(self, value):
+    def temperature_units(self, value: int):
         self.set_property(Props.TEMP_UNIT, int(value))
 
     @property
     def fan_speed(self) -> int:
-        return int(self.get_property(Props.FAN_SPEED))
+        return self.get_property(Props.FAN_SPEED)
 
     @fan_speed.setter
-    def fan_speed(self, value):
+    def fan_speed(self, value: int):
         self.set_property(Props.FAN_SPEED, int(value))
 
     @property
@@ -196,7 +297,7 @@ class Device:
         return bool(self.get_property(Props.FRESH_AIR))
 
     @fresh_air.setter
-    def fresh_air(self, value):
+    def fresh_air(self, value: bool):
         self.set_property(Props.FRESH_AIR, int(value))
 
     @property
@@ -204,7 +305,7 @@ class Device:
         return bool(self.get_property(Props.XFAN))
 
     @xfan.setter
-    def xfan(self, value):
+    def xfan(self, value: bool):
         self.set_property(Props.XFAN, int(value))
 
     @property
@@ -212,7 +313,7 @@ class Device:
         return bool(self.get_property(Props.ANION))
 
     @anion.setter
-    def anion(self, value):
+    def anion(self, value: bool):
         self.set_property(Props.ANION, int(value))
 
     @property
@@ -220,7 +321,7 @@ class Device:
         return bool(self.get_property(Props.SLEEP))
 
     @sleep.setter
-    def sleep(self, value):
+    def sleep(self, value: bool):
         self.set_property(Props.SLEEP, int(value))
 
     @property
@@ -228,31 +329,31 @@ class Device:
         return bool(self.get_property(Props.LIGHT))
 
     @light.setter
-    def light(self, value):
+    def light(self, value: bool):
         self.set_property(Props.LIGHT, int(value))
 
     @property
     def horizontal_swing(self) -> int:
-        return int(self.get_property(Props.SWING_HORIZ))
+        return self.get_property(Props.SWING_HORIZ)
 
     @horizontal_swing.setter
-    def horizontal_swing(self, value):
+    def horizontal_swing(self, value: int):
         self.set_property(Props.SWING_HORIZ, int(value))
 
     @property
     def vertical_swing(self) -> int:
-        return int(self.get_property(Props.SWING_VERT))
+        return self.get_property(Props.SWING_VERT)
 
     @vertical_swing.setter
-    def vertical_swing(self, value):
+    def vertical_swing(self, value: int):
         self.set_property(Props.SWING_VERT, int(value))
 
     @property
     def quiet(self) -> bool:
-        return bool(self.get_property(Props.QUIET))
+        return self.get_property(Props.QUIET)
 
     @quiet.setter
-    def quiet(self, value):
+    def quiet(self, value: bool):
         self.set_property(Props.QUIET, int(value))
 
     @property
@@ -260,7 +361,7 @@ class Device:
         return bool(self.get_property(Props.TURBO))
 
     @turbo.setter
-    def turbo(self, value):
+    def turbo(self, value: bool):
         self.set_property(Props.TURBO, int(value))
 
     @property
@@ -268,7 +369,7 @@ class Device:
         return bool(self.get_property(Props.STEADY_HEAT))
 
     @steady_heat.setter
-    def steady_heat(self, value):
+    def steady_heat(self, value: bool):
         self.set_property(Props.STEADY_HEAT, int(value))
 
     @property
@@ -276,6 +377,5 @@ class Device:
         return bool(self.get_property(Props.POWER_SAVE))
 
     @power_save.setter
-    def power_save(self, value):
+    def power_save(self, value: bool):
         self.set_property(Props.POWER_SAVE, int(value))
-
